#--*-perl-*-------------------------------------------------------------
#  input preparation for CNS scripts
#-----------------------------------------------------------------------
#$Author: saulius $
#$Date: 2008-03-13 15:48:36 +0200 (Thu, 13 Mar 2008) $
#$Revision: 1187 $
#$URL: svn+ssh://saulius-grazulis.lt/home/saulius/svn-repositories/cce/modules/CNS.pm $
#-----------------------------------------------------------------------

package CNS;

use strict;
require Exporter;
@CNS::ISA = qw(Exporter);
@CNS::EXPORT = qw(  );

use PDB;

my $HOME = $ENV{HOME};

$CNS::libs = defined $ENV{HOME} ? "$ENV{HOME}/lib/CNS" : "./";

#------------------------------------------------------------------------------
# read_pdb()
#
# Returns array of hashes, one hash for every .pdb file.
# Each hash can contain following fields (i.e. hash elements):
#
#    filename       pdb file name
#    topology       list of topology files
#    parameters     list of parameter files
#
#    rename         list of CNS renaming commands, to be
#                   executed after reading topology but
#                   before reading coordinates
#
#    commands       list of CNS commands, to be executed
#                   after reading topology, coordinates
#                   and reflections
#
#    segments       segment names encountered in this file
#    chains         chain names encountered in this file
#    interaction    selections of interacting atoms
#    fixed          selections of fixed atoms
#
#    linkfile       CNS .link file for this pdb file;
#                   empty string means that no link file
#                   should be used
#
#    alternative_mainchain
#    alternative_sidechains
#
#    atoms          a hash-of-hashes that contain atom records.
#                   The sub-hashes contain fields 'chains' and
#                   'alternatives'.

#
# The 'atoms' field should have the following structure:
#
# $pdb_descriptions[$i]{atoms} = {
#    chains => {
#        "A" => [ # chain A, alternatives "A" and " "
#            [ "ATOM  ", ... ] # segids are set to "A" (i.e. == chain id)
#            [ "ATOM  ", ... ]
#            [ "ATOM  ", ... ]
#        ],
#        "B" => [ # chain B, alternatives "A" and " "
#            [ "ATOM  ", ... ] # segids are set to "B" (i.e. == chain id)
#            [ "ATOM  ", ... ]
#            [ "ATOM  ", ... ]
#        ],
#    },
#    alternatives = {
#        "AB" => [ # chain A, alternative B
#            [ "ATOM  ", ... ] # segids are already set to "AB"
#            [ "ATOM  ", ... ]
#            [ "ATOM  ", ... ]
#        ],
#        "BB" => [ # chain B, alternative B
#            [ "ATOM  ", ... ] # segids are already set to "BB"
#            [ "ATOM  ", ... ]
#            [ "ATOM  ", ... ]
#        ],
#        "BC" => [ # chain B, alternative C
#            [ "ATOM  ", ... ] # segids are already set to "BC"
#            [ "ATOM  ", ... ]
#            [ "ATOM  ", ... ]
#        ],
#    },
# }
#

my @pdb_remark_fields = (
    "INTERACTION",
    "TOPOLOGY",
    "LINKFILE",
    "PARAMETERS",
    "FIXED",
    "ALTERNATIVE\\s+SIDECHAINS",
    "ALTERNATIVE\\s+MAINCHAIN",
    "DNA\\s+CHAINS",
    "RENAME",
    "COMMAND"
);

$::need_all_segments = 1;

sub read_pdb_segments { $::need_all_segments = 1; &read_pdb_all_or_first }
sub read_pdb          { $::need_all_segments = 0; &read_pdb_all_or_first }

sub read_pdb_all_or_first
{
    my @input_pdb = @_;
    my @pdb_description = ();

    my $keyword = join( "|", @pdb_remark_fields );
    my $file;
    foreach $file ( @input_pdb ) {
	my %pdb = ();
	my %segments = ();
	my %chains = ();
	if( ! -f $file ) {
	    warn( "file '$file' does not exist" );
	} else {
	    open( PDB, "$file" ) or
		die( "could not open '$file' for reading: $!" );

	    local $_;
	    while( <PDB> ) {
		if( /^ATOM  |^HETATM/ ) {
		    chomp;
		    my @atom = PDB::split_atom( $_ );

		    my $segname = $atom[14];
		    ## if( length($_) > 74 ) {
		    ## 	if( length($_) > 76 ) {
		    ## 	    $segname = substr( $_, 72, 4 );
		    ## 	} else {
		    ## 	    $segname = substr( $_, 72, length($_) - 74 );
		    ## 	}
		    ## } else {
		    ## 	$segname = "";
		    ## }
		    $segname =~ s/\s+$//g;
		    $segname = "\"$segname\"" if $segname =~ /\s/;
		    $segments{$segname} = $segname;
		    ## my $chname = substr( $_, 21, 1 );
		    my $chname = $atom[5];
		    $chains{$chname} = $chname;
		    unless( defined $pdb{first_atom} ) {
			$pdb{first_atom} = $_;
			chomp( $pdb{first_atom} );
		    }
		    ## my $resnr = substr( $_, 22, 4 );
		    my $resnr = $atom[6];
		    if( !defined $pdb{first_residue} ||
			substr( $pdb{first_residue}[0], 22, 4 ) == $resnr ) {
			push( @{$pdb{first_residue}}, $_ );
		    }

		    my $altpos = $atom[3];
		    if( $altpos eq " " || $altpos eq "A" ) {
			$atom[14] = $chname; # New segid
			push( @{$pdb{atoms}{chains}{$chname}}, \@atom );
		    } else {
			$atom[14] = $chname.$altpos; # New alt segid
			push( @{$pdb{atoms}{alternatives}{$chname.$altpos}},
			      \@atom );
		    }

		    $::need_all_segments ? next : last;
		}
		if( /^REMARK\s+CNS\s+NO\s+($keyword)/ ) {
		    my $key = "NO_$1";
		    $key =~ s/\s+/_/g;
		    $pdb{$key} = 1;
		    next;
		}
		if( /^REMARK\s+CNS\s+($keyword)\s+(.+)/ ) {
		    my $key = $1;
		    my $value = $2;
		    $key =~ s/\s+/_/g;
		    $value =~ s/^\s+|\s+$//g;
		    push( @{$pdb{$key}}, $value ) if $value;
		    next;
		}
	    }
	    close( PDB );
	}

	@{$pdb{segments}} = keys %segments;
	@{$pdb{chains}} = keys %chains;
	$pdb{filename} = $file;

	if( exists $pdb{atoms}{alternatives} ) {
	    my @alternatives = keys %{$pdb{atoms}{alternatives}};
	    my %alternative_sidechains = ();
	    for my $alt_segid ( sort {$a cmp $b} @alternatives ) {
		my $main_chain = substr( $alt_segid, 0, 1 );
		if( exists $alternative_sidechains{$main_chain} ) {
		    $alternative_sidechains{$main_chain} .= " ${alt_segid}";
		} else {
		    $alternative_sidechains{$main_chain} = ${alt_segid};
		}
	    }
	    for my $seg (sort {$a cmp $b} keys %alternative_sidechains ) {
		push( @{$pdb{ALTERNATIVE_SIDECHAINS}},
		      $seg . " " . $alternative_sidechains{$seg} );
	    }
	}

	push( @pdb_description, \%pdb );
    }

    do {
	use ShowStruct;
	showRef( \@pdb_description );
    } if 0;

    pdb_rename_dna( \@pdb_description );

    return @pdb_description;

} # read_pdb

#------------------------------------------------------------------------------

my %dna_rename = (
    "  A" => "ADE",
    "  C" => "CYT",
    "  G" => "GUA",
    "  T" => "THY",
    "  U" => "URI",

    "A  " => "ADE",
    "C  " => "CYT",
    "G  " => "GUA",
    "T  " => "THY",
    "U  " => "URI",

    " Ad" => "ADE",
    " Cd" => "CYT",
    " Gd" => "GUA",
    " Td" => "THY",
    " Ud" => "URI",

    " AD" => "ADE",
    " CD" => "CYT",
    " GD" => "GUA",
    " TD" => "THY",
    " UD" => "URI",

    " Ar" => "ADE",
    " Cr" => "CYT",
    " Gr" => "GUA",
    " Tr" => "THY",
    " Ur" => "URI",

    " AR" => "ADE",
    " CR" => "CYT",
    " GR" => "GUA",
    " TR" => "THY",
    " UR" => "URI",

    "Ad " => "ADE",
    "Cd " => "CYT",
    "Gd " => "GUA",
    "Td " => "THY",
    "Ud " => "URI",

    "AD " => "ADE",
    "CD " => "CYT",
    "GD " => "GUA",
    "TD " => "THY",
    "UD " => "URI",

    "Ar " => "ADE",
    "Cr " => "CYT",
    "Gr " => "GUA",
    "Tr " => "THY",
    "Ur " => "URI",

    "AR " => "ADE",
    "CR " => "CYT",
    "GR " => "GUA",
    "TR " => "THY",
    "UR " => "URI",
);

my %dna_atom_rename = (
    " C1*" => " C1'",
    " C2*" => " C2'",
    " C3*" => " C3'",
    " C4*" => " C4'",
    " C5*" => " C5'",
    " O2*" => " O2'",
    " O3*" => " O3'",
    " O4*" => " O4'",
    " O5*" => " O5'",
    " C5M" => " C5A",
);

sub pdb_rename_dna
{
    my $pdb_description = $_[0];

    for my $pdb (@{$pdb_description}) {
	for my $chain (keys %{$pdb->{atoms}{chains}}) {
	    my $atoms = $pdb->{atoms}{chains}{$chain};
	    for my $atom (@{$atoms}) {
		my $resname = $atom->[4];
		if( exists $dna_rename{$resname} ) {
		    $atom->[4] = $dna_rename{$resname};
		}
		my $atomname = $atom->[2];
		if( exists $dna_atom_rename{$atomname} ) {
		    $atom->[2] = $dna_atom_rename{$atomname};
		}
	    }
	}
	if( exists $pdb->{atoms}{alternatives} ) {
	    for my $segment (keys %{$pdb->{atoms}{alternatives}}) {
		my $atoms = $pdb->{atoms}{alternatives}{$segment};
		for my $atom (@{$atoms}) {
		    my $resname = $atom->[4];
		    if( exists $dna_rename{$resname} ) {
			$atom->[4] = $dna_rename{$resname};
		    }
		    my $atomname = $atom->[2];
		    if( exists $dna_atom_rename{$atomname} ) {
			$atom->[2] = $dna_atom_rename{$atomname};
		    }
		}
	    }
	}
    }
}

my %amino_acids = (
    "GLY" => "G",
    "ALA" => "A",
    "VAL" => "V",
    "LEU" => "L",
    "ILE" => "I",
    "PHE" => "F",
    "TYR" => "Y",
    "PRO" => "P",
    "SER" => "S",
    "THR" => "T",
    "CYS" => "C",
    "MET" => "M",
    "TRP" => "W",
    "ASP" => "D",
    "ASN" => "N",
    "GLU" => "E",
    "GLN" => "Q",
    "ARG" => "R",
    "LYS" => "K",
    "HIS" => "H",
);

my %dna_bases = (
    "ADE" => "A",
    "CYT" => "C",
    "GUA" => "G",
    "THY" => "T",
    "URI" => "U",

    "A"   => "A",
    "C"   => "C",
    "G"   => "G",
    "T"   => "T",
    "U"   => "U",

    "Ad"  => "A",
    "Cd"  => "C",
    "Gd"  => "G",
    "Td"  => "T",
    "Ud"  => "U",

    "Ar"  => "A",
    "Cr"  => "C",
    "Gr"  => "G",
    "Tr"  => "T",
    "Ur"  => "U",

    "AD"  => "A",
    "CD"  => "C",
    "GD"  => "G",
    "TD"  => "T",
    "UD"  => "U",

    "AR"  => "A",
    "CR"  => "C",
    "GR"  => "G",
    "TR"  => "T",
    "UR"  => "U",
);

my %water = (
    "HOH" => "O",
    "WAT" => "O",
);

my %ions = (
    "F"  => "F",
    "NA" => "NA",
    "MG" => "MG",
    "AL" => "AL",
    "CL" => "CL",
    "K"  => "K",
    "AR" => "AR",
    "CA" => "CA",
    "V"  => "V",
    "CR" => "CR",
    "MN" => "MN",
    "FE" => "FE",
    "NI" => "NI",
    "CO" => "CO",
    "CU" => "CU",
    "ZN" => "ZN",
    "AS" => "AS",
    "BR" => "BR",
    "KR" => "KR",
    "SR" => "SR",
    "MO" => "MO",
    "AG" => "AG",
    "CD" => "CD",
    "I"  => "I",
    "XE" => "XE",
    "CS" => "CS",
    "HO" => "HO",
    "YB" => "YB",
    "OS" => "OS",
    "IR" => "IR",
    "PT" => "PT",
    "AU" => "AU",
    "HG" => "HG",
    "PB" => "PB",
    "U"  => "U",
    "LI1" => "LI1",
    "F1"  => "F1",
    "NA1" => "NA1",
    "MG2" => "MG2",
    "AL3" => "AL3",
    "CL1" => "CL1",
    "K1"  => "K1",
    "CA2" => "CA2",
    "V2"  => "V2",
    "V3"  => "V3",
    "CR2" => "CR2",
    "CR3" => "CR3",
    "MN2" => "MN2",
    "MN3" => "MN3",
    "FE2" => "FE2",
    "FE3" => "FE3",
    "NI2" => "NI2",
    "CO2" => "CO2",
    "CO3" => "CO3",
    "CU1" => "CU1",
    "CU2" => "CU2",
    "ZN2" => "ZN2",
    "BR1" => "BR1",
    "SR2" => "SR2",
    "MO3" => "MO3",
    "AG1" => "AG1",
    "CD2" => "CD2",
    "I1"  => "I1",
    "CS1" => "CS1",
    "HO3" => "HO3",
    "YB2" => "YB2",
    "YB3" => "YB3",
    "OS4" => "OS4",
    "IR3" => "IR3",
    "PT2" => "PT2",
    "AU1" => "AU1",
    "AU3" => "AU3",
    "HG1" => "HG1",
    "HG2" => "HG2",
    "PB2" => "PB2",
    "U3"  => "U3",
    "U4"  => "U4",
    "SO4" => "SO4",
    "PO4" => "PO4",
    "WO4" => "WO4",
);

#
# sub guess_top_par_from_residue()
#
# Gets a name of a residue
#
# Returns array of three elements:
#
# ( "top-file-name.top", "par-file-name.param", "link-file-name.lnk" )
#
# Any of the three elements can be undef if not known or irrelevant.
#

sub guess_top_par_from_residue
{
    my $residue = $_[0];

    $residue =~ s/\s*|\s*$//g;

    if( defined $amino_acids{$residue} ) {
	return ( "CNS_TOPPAR:protein.top",
		 "CNS_TOPPAR:protein_rep.param",
		 "CNS_TOPPAR:protein.link" );
    }
    elsif( defined $dna_bases{$residue} ) {
	return ( "CNS_TOPPAR:dna-rna.top",
		 "CNS_TOPPAR:dna-rna_rep.param",
		 "CNS_TOPPAR:dna-rna.link" );
    }
    elsif( defined $water{$residue} ) {
	return ( "CNS_TOPPAR:water.top",
		 "CNS_TOPPAR:water.param",
		 undef );
    }
    elsif( defined $ions{$residue} ) {
	return ( "CNS_TOPPAR:ion.top",
		 "CNS_TOPPAR:ion.param",
		 undef );
    } else {
	return (undef, undef, undef);
    }
}

sub choose_default_toppar
{
    my @pdb_description = @_;

    for (@pdb_description) {
	if( defined $_->{NO_LINKFILE} ) {
	    @{$_->{LINKFILE}} = ( "" );
	}

	if( defined $_->{first_atom} ) {
	    my $first_residue = substr( $_->{first_atom}, 17, 3 );
	    my ( $top, $par, $link ) =
		guess_top_par_from_residue( $first_residue );

	    push( @{$_->{TOPOLOGY}}, $top ) if !$_->{TOPOLOGY} && $top;
	    push( @{$_->{PARAMETERS}}, $par ) if !$_->{PARAMETERS} && $par;
	    push( @{$_->{LINKFILE}}, $link ) if !$_->{LINKFILE} && $link;
	}
    }
    return @pdb_description;
}

sub choose_default_toppar_for_all_chains
{
    my @pdb_description = @_;

    for my $pdb (@pdb_description) {
	for my $chain (sort {$a cmp $b} keys %{$pdb->{atoms}{chains}}) {

	    my $atoms = $pdb->{atoms}{chains}{$chain};

	    if( defined $pdb->{NO_LINKFILE} ) {
		@{$pdb->{LINKFILE}} = ( "" );
	    } else {
		my $first_residue = $atoms->[0][4];
		my ( $top, $par, $link ) =
		    guess_top_par_from_residue( $first_residue );
		
		push( @{$pdb->{TOPOLOGY}}, $top ) if $top;
		push( @{$pdb->{PARAMETERS}}, $par ) if $par;
		push( @{$pdb->{LINKFILE}}, $link ) if $link;
	    }
	}
    }
    return @pdb_description;
}

#------------------------------------------------------------------------------

sub read_fob
{
    my @input_fob = @_;
    my @fob_description = ();

    my $file;
    foreach $file ( @input_fob ) {
	my %fob = ();
	if( ! -f $file ) {
	    warn( "file '$file' does not exist" );
	} else {
	    open( FOB, "$file" ) or
		die( "could not open '$file' for reading: $!" );

	    while( <FOB> ) {
		if( !/^\s*REMARK/i ) {
		    last;
		}
		if( /^\s*REMARK\s+RESOLUTION\s+(.+)$/i ) {
		    my $resolution = $1;
		    $resolution =~ s/^\s+|\s+$//g;
		    $fob{RESOLUTION} = $resolution;
		}
		if( /^\s*REMARK\s+CNS\s+COMMAND\s+(.+)$/i ) {
		    my $command = $1;
		    $command =~ s/^\s+|\s+$//g;
		    push( @{$fob{COMMAND}}, $command );
		}
	    }
	    close( FOB );
	}

	$fob{filename} = $file;
	push( @fob_description, \%fob );
    }

    return @fob_description;

} # read_fob

#------------------------------------------------------------------------------

sub emit_read_topology
{
    my $CNS = shift;
    my @pdb_description = @_;

    my %topology = ();

    for (@pdb_description) {
	if( $_->{TOPOLOGY} ) {
	    for( @{$_->{TOPOLOGY}} ) {
		$topology{$_} = $_;
	    }
	}
    }

    my @topology = keys %topology;
    if( @topology ) {
	my $top_file;
	foreach $top_file ( @topology ) {
	    print $CNS "topology \@\@$top_file end\n";
	}
	print $CNS "\n";
    }
}

sub emit_read_parameters
{
    my $CNS = shift;
    my @pdb_description = @_;

    my %parameters = ();

    for (@pdb_description) {
	if( $_->{PARAMETERS} ) {
	    for( @{$_->{PARAMETERS}} ) {
		$parameters{$_} = $_;
	    }
	}
    }

    my @parameters = sort keys %parameters;
    if( @parameters ) {
	my $par_file;
	foreach $par_file ( @parameters ) {
	    print $CNS "parameter \@\@$par_file end\n";
	}
	print $CNS "\n";
    }
}

sub emit_read_toppar_dir
{
    my $CNS = $_[0];
    my $toppar_dir = $_[1];
    my $pdb_descriptions = $_[2];
    my $pdb = defined $pdb_descriptions ? $pdb_descriptions->[0] : undef;

    $toppar_dir =~ s,/$,,;

    if( -d $toppar_dir ) {
	my @top_files = <${toppar_dir}/*.top>;
	my @par_files = <${toppar_dir}/*.par>;
	for (@top_files) {
	    print $CNS "topology \@\@$_ end\n";
	    if( defined $pdb ) {
		push( @{$pdb->{TOPOLOGY}}, $_ );
	    }
	}
	print $CNS "\n";
	for (@par_files) {
	    print $CNS "parameter \@\@$_ end\n";
	    if( defined $pdb ) {
		push( @{$pdb->{PARAMETERS}}, $_ );
	    }
	}
	print $CNS "\n";
    }
}

#------------------------------------------------------------------------------

$CNS::var_basename = "INPUT_PDB";

sub emit_define_pdb_vars
{
    local $CNS::var_basename = "INPUT_PDB";
    &emit_define_vars
}

sub emit_define_fob_vars
{
    local $CNS::var_basename = "INPUT_HKL";
    &emit_define_vars
}

sub emit_define_vars
{
    my $CNS = shift;
    my @descriptions = @_;

    if( @descriptions == 1 ) {
	my $input_name = "${CNS::var_basename}";
	print $CNS
	    "define ( $input_name = \"$descriptions[0]->{filename}\" )\n";
	$descriptions[0]{var_name} = $input_name;
    } else {
	my $count = 1;
	my $pdb_file;
	for ( @descriptions ) {
	    my $input_name = "${CNS::var_basename}_$count";
	    print $CNS "define ( $input_name = \"$_->{filename}\" )\n";
	    $_->{var_name} = $input_name;
	    $count ++;
	}
    }
}

#------------------------------------------------------------------------------

sub print_pdb_atoms
{
    my ($CNS, $atoms, $prefix ) = @_;

    my $format = "%s" . $PDB::atom_format;

    for my $atom (@$atoms) {
	printf $CNS $format, $prefix, @$atom;
    }
}

sub emit_read_chains
{
    my $CNS = shift;
    my $convert = shift;
    my $separate = shift;
    my @pdb_description = @_;
    my $convert_word = $convert ? "true" : "false";
    my $separate_word = $separate ? "true" : "false";

    local $\ = "\n";
    local $, = "\n";
    my $pdb;
    foreach  $pdb ( @pdb_description ) {
	my $segname = 
	    defined $pdb->{segments}[0] ? 
	    (
	     $pdb->{segments}[0] =~ /\".*\"/ ?
	     $pdb->{segments}[0] : "\"$pdb->{segments}[0]\""
	    ) : '""';

	if( ! exists $pdb->{atoms}{alternatives} ) {
	    print $CNS "segment";
	    print $CNS "   name=$segname" unless $convert or $separate;
	    print $CNS "   chain";
	    print $CNS "	 convert=$convert_word";
	    print $CNS "	 separate=$separate_word";
	    print $CNS "	 \@\@$pdb->{LINKFILE}[0]"
		if $pdb->{LINKFILE}[0];
	    if( defined $pdb->{var_name} ) {
		print $CNS "	 coordinates \@\@&$pdb->{var_name}";
	    } else {
		print $CNS "	 coordinates \@\@$pdb->{filename}";
	    }
	    print $CNS "   end";
	    print $CNS "end\n";
	} else {
	    print $CNS "segment";
	    print $CNS "   chain";
	    print $CNS "	 convert=false";
	    print $CNS "	 separate=true";
	    print $CNS "	 \@\@$pdb->{LINKFILE}[0]"
		if $pdb->{LINKFILE}[0];
	    print $CNS "	 coordinates";
	    for my $chain (sort {$a cmp $b} keys %{$pdb->{atoms}{chains}} ) {
		my $atoms = $pdb->{atoms}{chains}{$chain};
		print_pdb_atoms( $CNS, $atoms, "" );
	    }
	    for my $altseg (sort {$a cmp $b}
			    keys %{$pdb->{atoms}{alternatives}} ) {
		my $atoms = $pdb->{atoms}{alternatives}{$altseg};
		print_pdb_atoms( $CNS, $atoms, "" );
	    }
	    print $CNS "END";
	    print $CNS "   end {chain}";
	    print $CNS "end {segment}\n";
	}
    }
    my $has_rename = 0;
    foreach $pdb ( @pdb_description ) {
	if( defined $pdb->{RENAME} ) {
	    print $CNS @{$pdb->{RENAME}};
	    $has_rename = 1;
	}
    }
    print $CNS "" if $has_rename;
}

sub emit_read_chains_auto_splitted
{
    my $CNS = shift;
    my @pdb_description = @_;

    local $\ = "\n";
    local $, = "\n";

    my $pdb;
    foreach $pdb ( @pdb_description ) {
	my %chains = ();

	%chains = %{$pdb->{atoms}{chains}} if exists $pdb->{atoms}{chains};

	for my $chain (sort {$a cmp $b} keys %chains) {
	    print $CNS "segment";
	    print $CNS "   chain";
	    print $CNS "	 convert=false";
	    print $CNS "	 separate=true";

	    my $atoms = $chains{$chain};
	    my $resname = $atoms->[0][4];
	    my ( $top, $par, $link_file ) =
		guess_top_par_from_residue( $resname );

	    print $CNS "	 \@\@$link_file" if $link_file;
	    print $CNS "	 coordinates";
	    print_pdb_atoms( $CNS, $atoms, "" );
	    for my $altseg (sort {$a cmp $b}
			    keys %{$pdb->{atoms}{alternatives}} ) {
		if( $altseg =~ /^$chain/ ) {
		    my $atoms = $pdb->{atoms}{alternatives}{$altseg};
		    print_pdb_atoms( $CNS, $atoms, "" );
		}
	    }
	    print $CNS "END";
	    print $CNS "   end {chain}";
	    print $CNS "end {segment}\n";
	}
    }
    my $has_rename = 0;
    foreach $pdb ( @pdb_description ) {
	if( defined $pdb->{RENAME} ) {
	    print $CNS @{$pdb->{RENAME}};
	    $has_rename = 1;
	}
    }
    print $CNS "" if $has_rename;
}

sub emit_patch_segments_rna_to_dna
{
    my $CNS = shift;
    my $segments = join( " or ", map { "segid $_" } @_ );
    local $\ = "\n";
    print $CNS ' evaluate ($counter=0)';
    print $CNS ' for $id in id ( tag and ('."$segments".') ) loop dna';
    print $CNS '   evaluate ($counter=$counter+1)';
    print $CNS '   show (segid) (id $id)';
    print $CNS '   evaluate ($dna.segid.$counter=$result)';
    print $CNS '   show (resid) (id $id)';
    print $CNS '   evaluate ($dna.resid.$counter=$result)';
    print $CNS ' end loop dna';
    print $CNS ' evaluate ($dna.num=$counter)';
    print $CNS '';
    print $CNS ' evaluate ($counter=0)';
    print $CNS ' while ($counter < $dna.num) loop dnap';
    print $CNS '   evaluate ($counter=$counter+1)';
    print $CNS '   patch deox reference=nil=(segid $dna.segid.$counter and';
    print $CNS ' 'x30,'resid $dna.resid.$counter) end';
    print $CNS ' end loop dnap';
    print $CNS '';
}

sub emit_auto_patch_rna_to_dna
{
    my ( $CNS, $convert, @pdb_description ) = @_;

    for my $pdb (@pdb_description) {
	my @chains;
	if( $_->{DNA_CHAINS} ) {
	    @chains = map {split} @{$pdb->{DNA_CHAINS}};
	} else {
	    for my $chain (keys %{$pdb->{atoms}{chains}}) {
		my $atoms = $pdb->{atoms}{chains}{$chain};
		my $resname = $atoms->[0][4];
		next unless defined $dna_bases{$resname};

		my $atomname = "";
	      ATOM:
		for my $atom (@{$atoms}) {
		    $atomname = $atom->[4];
		    if( $atomname eq " O2'" or $atomname eq " O2*" ) {
			last ATOM;
		    }
		}
		if( $atomname ne " O2'" and $atomname ne " O2*" ) {
		    push( @chains, $convert ? $chain : $atoms->[0][14] );
		}
	    }
	}
	if( @chains ) {
	    emit_patch_segments_rna_to_dna( $CNS, @chains );
	}
    }
}

sub emit_read_coordinates
{
    my $CNS = shift;
    my $convert = shift;
    my @pdb_description = @_;

    $convert = $convert ? "true" : "false";

    foreach my $pdb ( @pdb_description ) {
	if( !exists $pdb->{atoms}{alternatives} ) {
	    if( defined $pdb->{var_name} ) {
		print $CNS "coordinates convert=${convert} \@&" .
		    "$pdb->{var_name}\n";
	    } else {
		print $CNS "coordinates convert=${convert} \@&" .
		    "$pdb->{filename}\n";
	    }
	} else {
	    print $CNS "coordinates convert=false\n";
	    for my $chain (sort {$a cmp $b} keys %{$pdb->{atoms}{chains}} ) {
		my $atoms = $pdb->{atoms}{chains}{$chain};
		print_pdb_atoms( $CNS, $atoms, "" );
	    }
	    for my $altseg (sort {$a cmp $b}
			    keys %{$pdb->{atoms}{alternatives}} ) {
		my $atoms = $pdb->{atoms}{alternatives}{$altseg};
		print_pdb_atoms( $CNS, $atoms, "" );
	    }
	    print $CNS "END\n";
	}
    }
    print $CNS "\n";
}

sub emit_read_coordinates_auto_splitted
{
    my $CNS = shift;
    my @pdb_description = @_;

    foreach my $pdb ( @pdb_description ) {
	my %chains = ();

	%chains = %{$pdb->{atoms}{chains}} if exists $pdb->{atoms}{chains};

	for my $chain (sort {$a cmp $b} keys %chains) {
	    print $CNS "coordinates convert=false\n";
	    my $atoms = $pdb->{atoms}{chains}{$chain};
	    print_pdb_atoms( $CNS, $atoms, "" );
	    for my $altseg (sort {$a cmp $b}
			    keys %{$pdb->{atoms}{alternatives}} ) {
		if( $altseg =~ /^$chain/ ) {
		    my $atoms = $pdb->{atoms}{alternatives}{$altseg};
		    print_pdb_atoms( $CNS, $atoms, "" );
		}
	    }
	    print $CNS "END\n";
	}
    }
    print $CNS "\n";
}

#------------------------------------------------------------------------------

$CNS::cell_variable = '$cell';
$CNS::sg_variable = '$sg';
$CNS::sgparam_variable = '$sgparam';

sub emit_read_reflections
{
    my $CNS = shift;
    my @fob_description = @_;

    return unless @fob_description;

    $fob_description[0]{cell_variable}    = $CNS::cell_variable;
    $fob_description[0]{sg_variable}      = $CNS::sg_variable;
    $fob_description[0]{sgparam_variable} = $CNS::sgparam_variable;

    my $access_name = defined $fob_description[0]->{var_name} ?
	"\&$fob_description[0]->{var_name}" : $fob_description[0]->{filename};

    print $CNS
        "\@${CNS::libs}/hklread(\n" .
        "        hklfile=${access_name};\n" .
	"        cell=$CNS::cell_variable;\n" .
        "        sg=$CNS::sg_variable; param=$CNS::sgparam_variable )\n\n";

    if( @fob_description > 1 ) {
	print $CNS "xray\n";
	my $last_fob_description = $#fob_description;
	for ( @fob_description[1..$last_fob_description] ) {
	    my $access_name = defined $_->{var_name} ?
		"\&$_->{var_name}" : $_->{filename};
	    print $CNS "    reflection \@\@${access_name} end\n";
	}
	print $CNS "end\n\n";
    }
}

#------------------------------------------------------------------------------

sub emit_select_reflections
{
    my ( $CNS, $hires, $lores, $fmin, $fmax, $sigmacut ) = @_;

print $CNS <<EOF;
\@${CNS::libs}/hklselect( hi_res=$hires; lo_res=$lores;
                        f_min=$fmin; f_max=$fmax;
                        sigma_cut=$sigmacut )

EOF
}


#------------------------------------------------------------------------------

sub emit_define_main_chain
{
    my ( $CNS ) = @_;

print $CNS <<EOF;
define ( main_chain = (name CA or name C or name N or name O))

EOF
}

sub emit_bfactor_reset_without_defining_main_chain
{
    my ( $CNS, $mainchain_b, $sidechain_b ) = @_;

print $CNS <<EOF;
do (b=$mainchain_b) (&main_chain)
do (b=$sidechain_b) (not &main_chain)

EOF
}

sub emit_bfactor_reset
{
    my ( $CNS, $mainchain_b, $sidechain_b ) = @_;

print $CNS <<EOF;
define ( main_chain = (name CA or name C or name N or name O))

do (b=$mainchain_b) (&main_chain)
do (b=$sidechain_b) (not &main_chain)

EOF
}


sub emit_occupancy_reset
{
    my $CNS = shift;

print $CNS <<EOF;
do (q=1.0) (attribute q = 0.0)

EOF
}

#------------------------------------------------------------------------------

sub emit_flags
{
    my $CNS = shift;
    my %flags = @_;

    print $CNS "flags\n";
    print $CNS "    exclude $flags{exclude}\n" if defined $flags{exclude};
    print $CNS "    include $flags{include}\n" if defined $flags{include};
    print $CNS "    ?\n";
    print $CNS "end\n\n";
}

#------------------------------------------------------------------------------

sub emit_setup_alt_chains
{
    my $CNS = shift;
    my @pdb_description = @_;

    my $echo_off = 0;

    for( @pdb_description ) {

	next unless
	    $_->{ALTERNATIVE_MAINCHAIN} ||
	    $_->{ALTERNATIVE_SIDECHAINS};

	unless( $echo_off ) {
	    print $CNS "set message=? end eval ( \$old_message = \$result )\n";
	    print $CNS "set echo=?    end eval ( \$old_echo    = \$result )\n";
	    print $CNS "set message=off echo=off end\n\n";
	    $echo_off = 1;
	}

	if( $_->{ALTERNATIVE_MAINCHAIN} ) {
	    my @alternative_chains = ();
	    for ( @{$_->{ALTERNATIVE_MAINCHAIN}} ) {
		push( @alternative_chains, [ split ] );
	    }
	    _emit_setup_alt_( $CNS, "main", @alternative_chains );
	}
	if( $_->{ALTERNATIVE_SIDECHAINS} ) {
	    my @alternative_chains = ();
	    for ( @{$_->{ALTERNATIVE_SIDECHAINS}} ) {
		push( @alternative_chains, [ split ] );
	    }
	    _emit_setup_alt_( $CNS, "side", @alternative_chains );
	}
    }
    print $CNS "set message=\$old_message echo=\$old_echo end\n\n"
	if( $echo_off );
}

sub _emit_setup_alt_
{
    my $CNS = shift;
    my $chain = shift;
    my @alternative_chains = @_;

    my $cns_module;

    if( $chain eq "side" )   { $cns_module = "setup_alt_sidechains" }
    elsif($chain eq "main" ) { $cns_module = "setup_alt_mainchain" }
    else { die }

    my $alt;
    for $alt (@alternative_chains) {
	my @alt = @$alt;
	my $nalt = int( @alt );
	my $main_seg = shift( @alt );
	my $alt_seg;
	for $alt_seg (@alt) {
	    my $new_alt_seg = "XALT";
	    print $CNS
		"\@${CNS::libs}/${cns_module}( \n" .
                "                   main_seg = \"$main_seg\";\n" .
		"                   alt_seg  = \"$alt_seg\";\n" .
		"                   new_alt_seg = \"$new_alt_seg\";\n" .
		"                   nalt = $nalt )\n\n";

	    print $CNS "delete select=(segid $alt_seg) end\n";
	    print $CNS "do (segid=\"$alt_seg\") (segid $new_alt_seg)\n\n";
	}
    }
}

#------------------------------------------------------------------------------

sub _emit_rename_alt_
{
    my $CNS = shift;
    my $count = shift;
    my $group_key = shift;
    my $pdb = shift;
    my @alternatives = @_;

    for ( @alternatives ) {
	my @alt = split( " ", $_ );
	my @group = ();
	my $old_main_seg = shift( @alt );
	my $new_main_seg = "${old_main_seg}${count}";

	push( @group, $old_main_seg );
	push( @group, $new_main_seg );
	$count ++;

	my $alt_seg;
	for $alt_seg (@alt) {
	    push( @group, $alt_seg );
	    print $CNS
		"\@${CNS::libs}/rename_main_alt(\n" .
                "              main_seg = \"$old_main_seg\";\n" .
		"              alt_seg  = \"$alt_seg\";\n" .
		"              new_main_seg = \"$new_main_seg\" )\n\n";
	}
	push( @{$pdb->{$group_key}}, \@group );
    }
    return $count;
}

sub emit_rename_alt_chains
{
    my $CNS = shift;
    my @pdb_description = @_;

    my $count = 1;
    my $echo_off = 0;

    my $pdb;
    for $pdb (@pdb_description) {

	if( ! defined $pdb->{ALTERNATIVE_MAINCHAIN} &&
	    ! defined $pdb->{ALTERNATIVE_SIDECHAINS} ) {
	    next;
	}

	unless( $echo_off ) {
	    print $CNS "set message=? end eval ( \$old_message = \$result )\n";
	    print $CNS "set echo=?    end eval ( \$old_echo    = \$result )\n";
	    print $CNS "set message=off echo=off end\n\n";
	    $echo_off = 1;
	}

	if( $pdb->{ALTERNATIVE_SIDECHAINS} ) {
	    $count = _emit_rename_alt_( $CNS, $count, "alt_side_groups",
					$pdb,
					@{$pdb->{ALTERNATIVE_SIDECHAINS}} );
	}

	if( $pdb->{ALTERNATIVE_MAINCHAIN} ) {
	    $count = _emit_rename_alt_( $CNS, $count, "alt_main_groups",
					$pdb,
					@{$pdb->{ALTERNATIVE_MAINCHAIN}} );
	}

	print $CNS "set message=\$old_message echo=\$old_echo end\n\n"
	    if $echo_off;
    }
}

#------------------------------------------------------------------------------

sub emit_back_rename_alt
{
    my $CNS = shift;
    my @pdb_description = @_;

    my $pdb;
    for $pdb (@pdb_description) {
	if( $pdb->{alt_side_groups} ) {
	    for ( @{$pdb->{alt_side_groups}} ) {
		my @group = @$_;
		my $old_main_seg = shift( @group );
		my $new_main_seg = shift( @group );
		print $CNS
		    "do (segid=\"$old_main_seg\") (segid $new_main_seg)\n";
	    }
	}

	if( $pdb->{alt_main_groups} ) {
	    for ( @{$pdb->{alt_main_groups}} ) {
		my @group = @$_;
		my $old_main_seg = shift( @group );
		my $new_main_seg = shift( @group );
		print $CNS
		    "do (segid=\"$old_main_seg\") (segid $new_main_seg)\n";
	    }
	}
    }
}

#------------------------------------------------------------------------------

sub get_fixed_atom_selections($$)
{
    my $pdb_description = $_[0];
    my $keys = $_[1];
    my @fixed;

    for my $pdb ( @$pdb_description ) {
	for my $key ( @$keys ) {
	    if( $pdb->{$key} ) {
		push( @fixed, @{$pdb->{$key}} );
	    }
	}
    }    

    return @fixed;
}

sub print_CNS_fixed_statement($$)
{
    my $CNS = $_[0];
    my $fixed = $_[1];

    if( $fixed && @$fixed ) {
	my @fixed = map { /^\(.*\)$/ ? $_ : "($_)" } @$fixed;
	my $fixed = join( " or ", @fixed );
	print $CNS "define( fixed = ($fixed) )\n\n";
    } else {
	print $CNS "define( fixed = (none) )\n\n";
    }
}

sub emit_define_fixed_user_only($@)
{
    my $CNS = shift;
    my @pdb_description = @_;

    my @fixed =
	get_fixed_atom_selections( \@pdb_description, [ "fixed" ] );

    print_CNS_fixed_statement( $CNS, \@fixed );
}

sub emit_define_fixed($@)
{
    my $CNS = shift;
    my @pdb_description = @_;


    my @fixed =
	get_fixed_atom_selections( \@pdb_description, [ "fixed", "FIXED" ] );

    print_CNS_fixed_statement( $CNS, \@fixed );
}

#------------------------------------------------------------------------------

sub emit_rename_commands
{
    my $CNS = shift;
    my @pdb_description = @_;

    my $pdb;
    foreach  $pdb ( @pdb_description ) {
	local $, = "\n";
	local $\ = "\n";
	if( defined $pdb->{rename_commands} ) {
	    print $CNS @{$pdb->{rename_commands}};
	}
    }
    print $CNS "\n";
}

sub emit_commands
{
    my $CNS = shift;
    my @pdb_description = @_;

    my $nl = "";
    my $pdb;
    foreach  $pdb ( @pdb_description ) {
	local $, = "\n";
	local $\ = "\n";
	if( defined $pdb->{COMMAND} ) {
	    print $CNS @{$pdb->{COMMAND}};
	    $nl = "\n";
	}
    }
    print $CNS $nl;
}

#------------------------------------------------------------------------------

sub emit_setup_interactions
{
    my $CNS = shift;
    my @pdb_description = @_;

    my %main_segments = ();
    my @alt_segments = ();
    my @interactions = ();
    my %has_interactions = ();
    my %segments = ();

    for ( @pdb_description ) {
	if( $_->{alt_side_groups} ) {
	    for( @{$_->{alt_side_groups}} ) {
		my @group = @$_;
		my $old_main_chain = shift( @group );
		$main_segments{$old_main_chain} = $old_main_chain;
		push( @alt_segments, @group );
	    }
	}
	if( $_->{alt_main_groups} ) {
	    for( @{$_->{alt_main_groups}} ) {
		my @group = @$_;
		my $old_main_chain = shift( @group );
		$main_segments{$old_main_chain} = $old_main_chain;
		push( @alt_segments, @group );
	    }
	}
	if( $_->{INTERACTION} ) {
	    for( @{$_->{INTERACTION}} ) {
		my $interaction = $_;
		push( @interactions, $interaction );
		while( $interaction =~ /segid\s+\"?([\w\s]+)/i ) {
		    $has_interactions{$1} = $1;
		    $interaction =~ s/segid\s+\"?([\w\s]+)//i;
		}
	    }
	}
	if( $_->{segments} ) {
	    for( @{$_->{segments}} ) {
		$segments{$_} = $_;
	    }
	}
    }

    my @main_segments = sort keys %main_segments;

    print $CNS "igroup\n";

    if( @main_segments || @alt_segments ) {
	my $select_main = "";
	if( @main_segments ) {
	    $select_main = join( " or ", map {"segid $_"} @main_segments );
	    print $CNS "    interaction ($select_main) ($select_main)\n";
	}
	my $alt;
	for $alt ( @alt_segments ) {
	    my $select_alt = "segid $alt";
	    print $CNS
		"    interaction ($select_main or $select_alt) " .
		    "($select_alt)\n";
	}
    }

    for ( @main_segments, @alt_segments ) {
	$has_interactions{$_} = $_;
    }

    for (@interactions) {
	print $CNS "    interaction $_\n";
    }

    for( keys %segments ) {
	unless ( $has_interactions{$_} ) {
	    if( $_ ne "" ) {
		$_ = "\"$_\"" if /^\s*$/;
		print $CNS "    interaction (segid $_ and known) (known)\n";
	    } else {
		print $CNS "    interaction (known) (known)\n";
		last;
	    }
	}
    }
    print $CNS "end\n\n";
}

#------------------------------------------------------------------------------

sub emit_refine_bfactors
{
    my $CNS = shift;

    &emit_refine_bfactors_indent( $CNS, "", @_ );
}

sub emit_refine_bfactors_indent
{
    my ( $CNS, $indent, $bmin, $bmax, $nsteps, $drop,
	 $bsig_main, $bsig_side, $asig_main, $asig_side ) = @_;

    print $CNS $indent, "xray\n";
    print $CNS $indent, "    optimise bfactor\n";
    print $CNS $indent, "        bmin=$bmin  bmax=$bmax\n";
    print $CNS $indent, "        nstep=$nsteps\n";
    print $CNS $indent, "        drop=$drop\n";
    print $CNS $indent, "        rweight=\$rweight_current\n";

    if( $bsig_main ) {
	print $CNS $indent, "        bsigma=(not(&fixed) and (&main_chain))=".
	                                    "$bsig_main\n";
    }
    if( $bsig_side ) {
	print $CNS $indent, "        bsigma=(not(&fixed) and " .
	                                    "not(&main_chain))=".
                                            "$bsig_side\n";
    }
    if( $asig_main ) {
	print $CNS $indent, "        asigma=(not(&fixed) and (&main_chain))=".
	                                    "$asig_main\n";
    }
    if( $asig_side ) {
	print $CNS $indent, "        asigma=(not(&fixed) and " .
	                                    "not(&main_chain))=".
					    "$asig_side\n";
    }
    print $CNS $indent, "        ?\n";
    print $CNS $indent, "    end {optimise bfactor}\n";
    print $CNS $indent, "end\n";
    print $CNS "\n";
}

#------------------------------------------------------------------------------

sub emit_refine_alt_occupancies
{
    my $CNS = shift;
    my $nsteps = shift;
    my @pdb_description = @_;

    my @alt_mainchain_groups = ();
    my @alt_sidechain_groups = ();

    for ( @pdb_description ) {
	if( $_->{alt_side_groups} ) {
	    for( @{$_->{alt_side_groups}} ) {
		my @group = @$_;
		shift( @group );
		push( @alt_sidechain_groups, \@group );
	    }
	}
	if( $_->{alt_main_groups} ) {
	    for( @{$_->{alt_main_groups}} ) {
		my @group = @$_;
		shift( @group );
		push( @alt_mainchain_groups, \@group );
	    }
	}
    }

    print $CNS "set message=? end eval ( \$old_message = \$result )\n";
    print $CNS "set echo=?    end eval ( \$old_echo    = \$result )\n";
    print $CNS "set message=off echo=off end\n\n";

    for ( @alt_sidechain_groups ) {
	my $segid = $_->[0];
	print $CNS "eval ( \$n$segid = 0 )\n";
	print $CNS "for \$i in id ( name CA and segid $segid ) loop alt\n";
	print $CNS "    show (resid) (id \$i)\n";
	print $CNS "    eval ( \$alt${segid}_\$n$segid = \$result )\n";
	print $CNS "    eval ( \$n$segid = \$n$segid + 1 )\n";
	print $CNS "end loop alt\n\n";
    }
    print $CNS "set message=\$old_message echo=\$old_echo end\n\n";

print $CNS <<EOF;
  xray
 	optimize group
 
 	   !! bmin=2.0  bmax=100.0
 	   nstep=$nsteps

EOF

    for ( @alt_sidechain_groups ) {
	my @alt = @$_;
	my $segid = shift( @alt );
	my $i = "           ";

	print $CNS "${i}eval ( \$count = 0 )\n";
	print $CNS "${i}while ( \$count < \$n${segid} ) loop alt\n";
	print $CNS
	    "${i}    q = (segid $segid and resid \$alt${segid}_\$count)\n";

	my $alt;
	for $alt (@alt) {
	    print $CNS
		"${i}    q = (segid $alt and resid \$alt${segid}_\$count)\n";
	}

	print $CNS "${i}    eval ( \$count = \$count + 1 )\n";
	print $CNS "${i}end loop alt\n\n";
    }

    for ( @alt_mainchain_groups ) {
	for (@$_) {
	    my $i = "           ";
	    print $CNS "${i}q = (segid $_)\n";
	}
    }
    print $CNS "\n" if @alt_mainchain_groups;

print $CNS <<EOF;
 	   ?
 	end
 
  end

EOF

print $CNS <<EOF;
!------------------------------------------------------------------------------
! ensure that the sum of alternative occupancies is 1.0

 set message=? end eval ( \$old_message = \$result )
 set echo=?    end eval ( \$old_echo    = \$result )
 set message=off echo=off end

EOF

    for (@alt_sidechain_groups, @alt_mainchain_groups) {
	my @alt = @$_;
	my $nalt = int(@alt);
	my $main_seg = $alt[0];

print $CNS <<EOF;

for \$i in id ( name CA and segid $main_seg ) loop alt
    show (resid) (id \$i)
    eval ( \$resid = \$result )
    display \$resid
    for \$atom in id ( segid $main_seg and resid \$resid ) loop coord
   	show (name) (id \$atom)
 	eval ( \$name = \$result )

EOF

        my @q = ();
	my $count = 1;
	my $i = "        ";
	for (@alt) {
	    print $CNS
		"${i}show (q) (segid $_ and name \$name and resid \$resid)\n";
	    print $CNS
		"${i}eval ( \$q${count} = \$result )\n";
	    push( @q, "\$q${count}" );
	    $count ++;
	}

	print $CNS "${i}eval ( \$sum = " . join( " + ", @q ) . " )\n";

	for (@alt) {
	    print $CNS
		"${i}do (q=q/\$sum) (segid $_ and " .
		    "name \$name and resid \$resid)\n";
	}
	print $CNS "\n";

	print $CNS "    end loop coord\n";
	print $CNS "end loop alt\n\n";
    }

    print $CNS "set message=\$old_message echo=\$old_echo end\n\n";
}

#------------------------------------------------------------------------------

$CNS::bcorrection_none             = "none";
$CNS::bcorrection_bulk_anisotropic = "bulk anisotropic";
$CNS::bcorrection_bulk_isotropic   = "bulk isotropic";
$CNS::bcorrection_anisotropic      = "anisotropic";
$CNS::bcorrection_isotropic        = "isotropic";

sub emit_setup_bcorrection
{
    my $CNS = shift;
    my $bcorrection = shift;

print $CNS <<EOF;
\@${CNS::libs}/provideobject( object=fpart;
     domain=reciprocal; type=complex )

\@${CNS::libs}/predictfcalc( atoms=(all); fcalc=fcalc )

EOF

if( $bcorrection eq $CNS::bcorrection_none ) {
    # do nothing
}
elsif( $bcorrection eq $CNS::bcorrection_bulk_anisotropic ) {
    print $CNS <<EOF
\@${CNS::libs}/bscorrect-aniso( atoms=(all); resolution=6.0;
                         fobs=fobs; sigma=sigma;
                         fcalc=fcalc; fpart=fpart;
                         used=used; test=test )

EOF
}
elsif( $bcorrection eq $CNS::bcorrection_bulk_isotropic ) {
    print $CNS <<EOF
\@${CNS::libs}/bscorrect-iso( atoms=(all); resolution=6.0;
                       fobs=fobs;
                       fcalc=fcalc; fpart=fpart;
                       used=used; test=test )

EOF
} else {
    die( "unknown b-factor correction type '$bcorrection'" );
}
}

#------------------------------------------------------------------------------

$CNS::target_residual = "residual";
$CNS::target_mlf      = "mlf";

sub emit_setup_target
{
    my $CNS = shift;
    my $target = shift;

if( $target eq $CNS::target_residual ) {
    # do nothing -- residual target is a CNS default
}
elsif( $target eq $CNS::target_mlf ) {
    print $CNS <<EOF
\@${CNS::libs}/setup_target_mlf( fobs=fobs; sigma=sigma;
                          fcalc=fcalc; fpart=fpart;
                          used=used; test=test )

EOF
} else {
    die( "unknown target '$target'" );
}
}

#------------------------------------------------------------------------------

sub emit_write_coordinates
{
    &emit_write_coordinates_indent( $_[0], "", $_[1] );
}

sub emit_write_coordinates_indent
{
    my $CNS = shift;
    my $indent = shift;
    my $filename = shift;

    print $CNS $indent, "print thres=0.02 bonds\n";
    print $CNS $indent, "print thres=5. angles\n\n";
    print $CNS $indent, "remark r= \$monitor free_r= \$test_monitor\n\n";
    print $CNS $indent, "\@${CNS::libs}/pdbwrite(\n";
    print $CNS $indent, "        output=${filename};\n";
    print $CNS $indent, "        cell=$CNS::cell_variable;\n";
    print $CNS $indent, "        sgparam=$CNS::sgparam_variable )\n";
    print $CNS "\n";
}

#------------------------------------------------------------------------------

sub restore_cns_remarks
{
    my $output_pdb = shift;
    my @pdb_description = @_;

    local $^I = "";
    local @ARGV = ( $output_pdb );

    my $header;

    my %topology = ();
    my %parameters = ();

    for ( @pdb_description ) {
	if( $_->{TOPOLOGY} ) {
	    my $topology;
	    for $topology (@{$_->{TOPOLOGY}}) {
		$topology{$topology} = "";
	    }
	}
	if( $_->{PARAMETERS} ) {
	    my $parameters;
	    for $parameters (@{$_->{PARAMETERS}}) {
		$parameters{$parameters} = "";
	    }
	}
    }

    local $_;
    while(<>) {
	unless( defined $header ) {
	    my $top_file;
	    foreach $top_file (sort keys %topology) {
		print "REMARK CNS TOPOLOGY $top_file\n";
	    }

	    my $par_file;
	    foreach $par_file (sort keys %parameters) {
		print "REMARK CNS PARAMETERS $par_file\n";
	    }

	    for (@pdb_description) {
		if( $_->{INTERACTION} ) {
		    my $interaction;
		    for $interaction (@{$_->{INTERACTION}}) {
			print "REMARK CNS INTERACTION $interaction\n";
		    }
		}
		if( $_->{FIXED} ) {
		    my $fixed;
		    for $fixed (@{$_->{FIXED}}) {
			print "REMARK CNS FIXED $fixed\n";
		    }
		}
		if( $_->{DNA_CHAINS} ) {
		    for ( @{$_->{DNA_CHAINS}} ) {
			print "REMARK CNS DNA CHAINS $_\n";
		    }
		}

		my $alt;
		local $" = " ";
		if( $_->{ALTERNATIVE_SIDECHAINS} ) {
		    for $alt (@{$_->{ALTERNATIVE_SIDECHAINS}}) {
			print "REMARK CNS ALTERNATIVE SIDECHAINS $alt\n";
		    }
		}
		if( $_->{ALTERNATIVE_MAINCHAIN} ) {
		    for $alt (@{$_->{ALTERNATIVE_MAINCHAIN}}) {
			print "REMARK CNS ALTERNATIVE MAINCHAIN @$alt\n";
		    }
		}
	    }
	    $header = 1;
	}

	print unless /^REMARK\s+CNS/;
    }
}

#------------------------------------------------------------------------------

1;
