#--*-perl-*-------------------------------------------------------------
#$Author: saulius $
#$Date: 2006-05-12 09:40:28 +0300 (Fri, 12 May 2006) $
#$Revision: 1056 $
#$URL: svn+ssh://saulius-grazulis.lt/home/saulius/svn-repositories/cce/modules/SymopAlg.pm $
#-----------------------------------------------------------------------
#*
# Basic symmetry operator algebra (addition, multiplication, etc.)
#**

package SymopAlg;

use strict;

require Exporter;
@SymopAlg::ISA = qw(Exporter);
@SymopAlg::EXPORT = qw( symop_translate symop_mul symop_inv symop_apply 
		        symop_apply_to_atom symop_apply_to_molecule );

sub matrix3x3_times_vector($$)
{
    my ( $m, $v ) = @_;
    return [
	    $m->[0][0] * $v->[0] + $m->[0][1] * $v->[1]	+ $m->[0][2] * $v->[2],
	    $m->[1][0] * $v->[0] + $m->[1][1] * $v->[1]	+ $m->[1][2] * $v->[2],
	    $m->[2][0] * $v->[0] + $m->[2][1] * $v->[1]	+ $m->[2][2] * $v->[2]
	   ];
}

sub matrix3x3_times_matrix($$)
{
    my ( $m1, $m2 ) = @_;
    my ( $i, $j );

    my @result;
    for( $i = 0; $i < 3; $i++ ) {
        for( $j = 0; $j < 3; $j++ ) {
            $result[$i][$j] = $m1->[$i][0] * $m2->[0][$j] +
		              $m1->[$i][1] * $m2->[1][$j] +
        		      $m1->[$i][2] * $m2->[2][$j];
        }
    }
    return \@result;
}

sub symop_translate($$)
{
    my ( $symop, $vector ) = @_;
    return [ [ @{$symop->[0]} ], [ @{$symop->[1]} ], [ @{$symop->[2]} ],
	     [
	      $symop->[3][0] + $vector->[0],
	      $symop->[3][1] + $vector->[1],
	      $symop->[3][2] + $vector->[2]
	     ]
	   ];
}

sub symop_mul($$)
{
    my ( $s1, $s2 ) = @_;

    my $rmat = matrix3x3_times_matrix( $s1, $s2 );
    my $vect = matrix3x3_times_vector( $s1, $s2->[3] );
    return [ [ @{$rmat->[0]} ], [ @{$rmat->[1]} ], [ @{$rmat->[2]} ],
	    [
	     $vect->[0] + $s1->[3][0],
	     $vect->[1] + $s1->[3][1],
	     $vect->[2] + $s1->[3][2]
	    ]];
}

sub matrix2x2_det($)
{
    my $m = $_[0];
    return $m->[0][0]*$m->[1][1] - $m->[0][1]*$m->[1][0];
}

sub symop_adjunct($$$)
{
    my ( $s, $row, $col ) = @_;
 
    my @matrix = ();
    my ( $i, $j, $mi, $mj );
    my $coef;

    die unless( $row >= 0 && $row < 3 );
    die unless( $col >= 0 && $col < 3 );

    $mi = $mj = 0;
    for( $i = 0; $i < 3; $i ++ ) {
        next if( $i == $row );
        $mj = 0;
        for( $j = 0; $j < 3; $j ++ ) {
	    next if( $j == $col );
            $matrix[$mi][$mj] = $s->[$i][$j];
            $mj ++;
        }
        $mi ++;
    }
    die unless( $mi == 2 );
    die unless( $mj == 2 );
    $coef = (($row + $col) % 2 == 0) ? +1.0 : -1.0;
    return $coef * matrix2x2_det( \@matrix );
}

sub symop_det( $ )
{
    my $s = $_[0];
    return
        + $s->[0][0] * $s->[1][1] * $s->[2][2]
        + $s->[1][0] * $s->[2][1] * $s->[0][2]
        + $s->[0][1] * $s->[1][2] * $s->[2][0]

        - $s->[0][2] * $s->[1][1] * $s->[2][0]
        - $s->[0][0] * $s->[1][2] * $s->[2][1]
        - $s->[0][1] * $s->[1][0] * $s->[2][2];
}

sub vector_negate($)
{
    return [ map {-$_} @{$_[0]} ];
}

sub symop_inv( $ )
{
    my $s = $_[0];
    my @ret;
    my $det = symop_det( $s );
    my ( $i, $j );

    for( $i = 0; $i < 3; $i++ ) {
        for( $j = 0; $j < 3; $j++ ) {
            $ret[$i][$j] = symop_adjunct($s,$j,$i) / $det;
        }
    }
    $ret[3] = vector_negate( matrix3x3_times_vector( \@ret, $s->[3] ));
    return \@ret;
}

sub symop_apply($$)
{
    my ( $symop, $vector ) = @_;
    $vector = matrix3x3_times_vector( $symop, $vector );
    return [
	    $vector->[0] + $symop->[3][0],
	    $vector->[1] + $symop->[3][1],
	    $vector->[2] + $symop->[3][2]
	   ];
}

sub symop_apply_to_atom($$)
{
    my ( $symop, $atom ) = @_;
    my $new_xyz = symop_apply( $symop, [@{$atom}[8,9,10]] );
    return [@{$atom}[0..7], @{$new_xyz}, @{$atom}[11..$#{$atom}]];
}

sub symop_apply_to_molecule($$)
{
    my ( $symop, $atoms ) = @_;
    my $new_atoms = [];
    my $atom;
    foreach $atom ( @$atoms ) {
	my $new_xyz = symop_apply( $symop, [@{$atom}[8,9,10]] );
	my $new_atom = [@{$atom}[0..7], @{$new_xyz}, @{$atom}[11..$#{$atom}]];
	push( @$new_atoms, $new_atom );
    }
    return $new_atoms;
}

1;
